"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.api.datasets import boston_housing
from keras.api.datasets import california_housing
from keras.api.datasets import cifar10
from keras.api.datasets import cifar100
from keras.api.datasets import fashion_mnist
from keras.api.datasets import imdb
from keras.api.datasets import mnist
from keras.api.datasets import reuters
